#include "pixelpipeline.h"
#include <framebuffer.h>
#include <texture.h>
#include <shader.h>
#include <cstring>
#include <glm/vec2.hpp>
#include <qtutils.h>
#include "mapscene.h"

#define SCROLL_SPEED 0.998f

#define HEX_TO_VEC3(hex) glm::vec3(float((hex & 0xFF0000) >> 16)/255, float((hex & 0x00FF00) >> 8)/255, float(hex & 0x0000FF)/255)

inline glm::vec3 getColor(const Pixel &p)
{
    switch(p.type){
		case Pixel::BEDROCK : return HEX_TO_VEC3(0x101020);
		case Pixel::GRASS : return HEX_TO_VEC3(0x719678);
		case Pixel::MARK : return HEX_TO_VEC3(0x5D7B62);
		case Pixel::ROCK : return HEX_TO_VEC3(0x8C8C8C);
		case Pixel::IRON_ORE : return HEX_TO_VEC3(0x917B61);
		case Pixel::TREE : return HEX_TO_VEC3(0x003800);
		case Pixel::BERRIES : return HEX_TO_VEC3(0x4D6394);
		case Pixel::FOOD : return HEX_TO_VEC3(0xFF7A7A);
		case Pixel::WOOD : return HEX_TO_VEC3(0x634A22);
		case Pixel::STONE : return HEX_TO_VEC3(0x454545);
		case Pixel::IRON : return HEX_TO_VEC3(0x4A4036);
		case Pixel::DUDE :
            // TODO
			return HEX_TO_VEC3(0x0000FF);
		case Pixel::SPAWN :
			// TODO
			return HEX_TO_VEC3(0x0000FF);
		case Pixel::WALL : return HEX_TO_VEC3(0xE6B2A1);
		case Pixel::ROAD : return HEX_TO_VEC3(0xEDB287);
		case Pixel::SWORD : return HEX_TO_VEC3(0xEBEBEB);
		case Pixel::LIBRARY : return HEX_TO_VEC3(0xA37A50);
		case Pixel::DEAD_DUDE : return HEX_TO_VEC3(0xFF0000);
		default : return HEX_TO_VEC3(0x0000FF); // bleu absolu = bug
	}
}


// MESH (2D points)

class PixelMesh
{
private:
   GLuint m_vao;
   GLuint m_vbo;
   
public:
    struct Pix
    {
        glm::vec2 pos;
        glm::vec3 color;
    };
    
    std::vector<Pix> m_pixels;
   
    PixelMesh(int size)
    {
        glGenVertexArrays(1, &m_vao);
        glBindVertexArray(m_vao);
        
        glGenBuffers(1, &m_vbo);
        glBindBuffer(GL_ARRAY_BUFFER, m_vbo);
        glEnableVertexAttribArray(0);
        glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, sizeof(Pix), (void*)(0));
        glEnableVertexAttribArray(1);
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, sizeof(Pix), (void*)(sizeof(glm::vec2)));
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        
        glBindVertexArray(0);
    }
   
    ~PixelMesh()
    {
        if(m_vao != 0)
        {
            glDeleteBuffers(1, &m_vbo);
            glDeleteVertexArrays(1, &m_vao);
            m_vao = 0;
        }
    }
   
   void draw()
   {
        glDisable(GL_BLEND);
        glDisable(GL_DEPTH_TEST);
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glBindVertexArray(m_vao);
        glBindBuffer(GL_ARRAY_BUFFER, m_vbo);
        glBufferData(GL_ARRAY_BUFFER, m_pixels.size() * sizeof(Pix), m_pixels.data(), GL_DYNAMIC_DRAW);
        glDrawArrays(GL_POINTS, 0, m_pixels.size());
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        glBindVertexArray(0);
        m_pixels.clear();
   }
};

// PIPELINE

PixelPipeline::PixelPipeline(MapScene *map) :
    m_mapWidth(map->getWidth()),
    m_mapHeight(map->getHeight()),
    m_camera(0, 0, 1)
{
    m_width = 256;
    m_height = 256;
    map->setPipeline(this);
    m_targetFBO = FrameBuffer::screen;
    m_mapFBO = new FrameBuffer();
    m_mapTex = new Texture(GL_RGBA, GL_RGBA, m_mapWidth, m_mapHeight, GL_UNSIGNED_BYTE, GL_TEXTURE_RECTANGLE);
    m_mapTex->setFiltering(GL_NEAREST);
    m_mapTex->setWrap(GL_CLAMP_TO_EDGE);
    m_mapFBO->addTexture(m_mapTex, GL_COLOR_ATTACHMENT0);
    m_mapFBO->initColorAttachments();

    // set up vao
    const float vertices[] = {
       -1.0f, -1.0f,
        3.0f, -1.0f,
       -1.0f,  3.0f
    };
    glGenVertexArrays(1, &m_vao);
    glBindVertexArray(m_vao);
    glGenBuffers(1, &m_vbo);
    glBindBuffer(GL_ARRAY_BUFFER, m_vbo);
    glBufferData(GL_ARRAY_BUFFER, 3 * 2 * sizeof(GLfloat), vertices, GL_STATIC_DRAW);
    glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, sizeof(GLfloat)*2, NULL);
    glEnableVertexAttribArray(0);
    
    std::string vertSource = QtUtils::fileToString(":shaders/shaders/pixel.vert.glsl").toStdString();
    std::string fragSource = QtUtils::fileToString(":shaders/shaders/pixel.frag.glsl").toStdString();
    m_texMapShader = new Shader(vertSource, fragSource);
    vertSource = QtUtils::fileToString(":shaders/shaders/world.vert.glsl").toStdString();
    fragSource = QtUtils::fileToString(":shaders/shaders/world.frag.glsl").toStdString();
    m_renderShader = new Shader(vertSource, fragSource);
    m_changes = new PixelMesh(map->getWidth() * map->getHeight());
    for(int i=0; i<map->getWidth(); ++i)
        for(int j=0; j<map->getHeight(); ++j)
            pushChange(glm::vec2(i, j), getColor((*map)[i][j]));
    updateChanges();
}

PixelPipeline::~PixelPipeline()
{
    delete m_mapFBO;
    delete m_mapTex;
    delete m_texMapShader;
    delete m_renderShader;
    delete m_changes;
}

void PixelPipeline::pushChange(const glm::vec2 &pos, const glm::vec3 &color)
{
    m_changes->m_pixels.push_back({pos, color});
}

void PixelPipeline::updateChanges()
{
    if(!m_changes->m_pixels.empty())
    {
        m_mapFBO->bindFBO();
        glViewport(0, 0, m_mapWidth, m_mapHeight);
        m_texMapShader->bind();
        m_texMapShader->bindVec2(m_texMapShader->getLocation("texRatio"), glm::vec2(2.0f/m_mapWidth, 2.0f/m_mapHeight));
        m_changes->draw();
    }
}

void PixelPipeline::renderGL(Scene *scene)
{
    m_targetFBO->bindFBO();
    glClearColor(0, 0, 0, 1);
    glClear(GL_COLOR_BUFFER_BIT);
    glViewport(0, 0, m_width, m_height);
    glDisable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);
    m_renderShader->bind();
    m_renderShader->bindVec3(m_renderShader->getLocation("camera"), m_camera);
    m_renderShader->bindVec2(m_renderShader->getLocation("worldSize"), glm::vec2(m_mapWidth, m_mapHeight));
    m_renderShader->bindVec2(m_renderShader->getLocation("screenSize"), glm::vec2(m_width, m_height));
    m_mapTex->bind(0);
    m_renderShader->bindInteger(m_renderShader->getLocation("colorMap"), 0);

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glBindVertexArray(m_vao);
    glDrawArrays(GL_TRIANGLES, 0, 3);
    glBindVertexArray(0);
}

void PixelPipeline::resizeGL(int w, int h)
{
    m_width = w;
    m_height = h;
}

void PixelPipeline::cameraZoom(int nbScrolls)
{
    while(nbScrolls != 0)
        {
            if(nbScrolls > 0)
            {
                m_camera.z *= SCROLL_SPEED;
                --nbScrolls;
            }
            else
            {
                m_camera.z /= SCROLL_SPEED;
                ++nbScrolls;
            }
        }
}
